import os
import html
import copy

from nbconvert import HTMLExporter, SlidesExporter


class Exporter:
    def __init__(
        self,
        show_code=False,
        show_prompt=False,
        is_presentation=False,
        reveal_theme="white",
    ):
        self.show_code = show_code
        self.show_prompt = show_prompt
        self.is_presentation = is_presentation
        self.reveal_theme = reveal_theme

        self.set_exporter()

    def set_exporter(self):
        if self.is_presentation:
            self.html_exporter = SlidesExporter(
                template_name="reveal", reveal_theme=self.reveal_theme
            )
        else:
            self.html_exporter = HTMLExporter(template_name="lab")

        # prompt
        self.html_exporter.exclude_input_prompt = not self.show_prompt
        self.html_exporter.exclude_output_prompt = not self.show_prompt

        # hide code
        self.html_exporter.exclude_input = not self.show_code
        self.html_exporter.exclude_input_prompt = not self.show_code
        self.html_exporter.exclude_output_prompt = not self.show_code

    def set_is_presentation(self, new_value):
        if self.is_presentation != new_value:
            self.is_presentation = new_value
            self.set_exporter()

    def set_reveal_theme(self, new_value):
        if self.reveal_theme != new_value:
            self.reveal_theme = new_value
            self.set_exporter()

    def set_show_code(self, new_value):
        self.show_code = new_value
        self.html_exporter.exclude_input = not self.show_code
        self.html_exporter.exclude_input_prompt = not self.show_code
        self.html_exporter.exclude_output_prompt = not self.show_code

    def set_show_prompt(self, new_value):
        self.show_prompt = new_value
        self.html_exporter.exclude_input_prompt = not self.show_prompt
        self.html_exporter.exclude_output_prompt = not self.show_prompt

    def is_stop_execution_output(self, output):
        return output.get(
            "output_type", ""
        ) == "error" and "StopExecution" in output.get("ename", "")

    def export(self, notebook, full_header=True):
        #
        # notebook needs to be in nbformat
        #
        n = copy.deepcopy(notebook)

        # omit cells with "---" (starter from V1)
        n.cells = [
            cell
            for cell in n.cells
            if not "".join(cell.get("source", "")).startswith("---")
        ]

        # omit output with mercury widgets
        for cell in n.cells:
            if "outputs" in cell:
                cell["outputs"] = [
                    output
                    for output in cell["outputs"]
                    if not "application/mercury+json" in output.get("data", {})
                ]
        # omit output with mercury StopExecution
        for cell in n.cells:
            if "outputs" in cell:
                cell["outputs"] = [
                    output
                    for output in cell["outputs"]
                    if not self.is_stop_execution_output(output)
                ]
        try:
            # replace iframe
            for cell in n.cells:
                if cell.get("outputs") is None:
                    continue
                for output in cell.outputs:
                    # check for iframe
                    any_html = output.get("data", {}).get("text/html", None)
                    if any_html is not None:
                        if isinstance(any_html, list):
                            any_html = "".join(any_html)
                        if "<iframe" in any_html:
                            place = any_html.find("src=")
                            if place != -1:
                                stop = any_html.find('"', place + 5)
                            if place != -1 and stop != -1:
                                fname = any_html[place + 5 : stop]

                                if not fname.startswith("http"):
                                    content = None
                                    if os.path.exists(fname):
                                        with open(
                                            fname,
                                            "r",
                                            encoding="utf-8",
                                            errors="ignore",
                                        ) as fin:
                                            content = html.escape(fin.read())
                                    if content is not None:
                                        any_html = any_html.replace(
                                            any_html[place : stop + 1],
                                            f'srcdoc="{content}"',
                                        )
                                        # new_output = [i+"\n" for i in any_html.split("\n")]
                                        output["data"][
                                            "text/html"
                                        ] = any_html  # new_output

        except Exception as e:
            pass

        body, _ = self.html_exporter.from_notebook_node(n)

        # remove header
        if not full_header:
            body = self.remove_header(body)

        return body

    def remove_header(self, nb_body):
        body = copy.deepcopy(nb_body)
        index_start = body.find("<head>")
        index_end = body.find("</head>")

        if index_start != -1 and index_end != -1:
            body = body[:index_start] + "" + body[index_end + 7 :]

            body = body.replace("<!DOCTYPE html>", "")
            body = body.replace("<html>", "")
            body = body.replace("<body ", "<div ")
            body = body.replace("</body>", "</div>")
            body = body.replace("</html>", "")
        return body



import logging

from django.db.models import Q
from django.http import JsonResponse
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Membership, Site
from apps.notebooks.models import Notebook
from apps.storage.models import UploadedFile, WorkerFile
from apps.storage.s3utils import S3
from apps.storage.serializers import UploadedFileSerializer
from apps.workers.models import Worker
from apps.accounts.views.utils import is_cloud_version

from apps.accounts.views.sites import (
    get_plan,
    PLAN_KEY,
    PLAN_STARTER,
    PLAN_PRO,
    PLAN_BUSINESS,
)

from apps.storage.utils import (
    get_bucket_key,
    get_site_bucket_key,
    get_worker_bucket_key,
)

log = logging.getLogger(__name__)


class WorkerPresignedUrl(APIView):
    def get(
        self,
        request,
        action,
        session_id,
        worker_id,
        notebook_id,
        output_dir,
        filename,
        format=None,
    ):
        try:
            # check if such worker exists
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            client_action = (
                "put_object" if action in ["put", "put_object"] else "get_object"
            )

            s3 = S3()
            url = s3.get_presigned_url(
                get_worker_bucket_key(
                    session_id, output_dir, filename.replace(" ", "-")
                ),
                client_action,
            )

            return Response({"url": url})

        except Exception as e:
            log.exception("Cant create presigned url for worker")

        return Response(status=status.HTTP_403_FORBIDDEN)


class WorkerAddFile(APIView):
    def post(self, request, format=None):
        worker_id = request.data.get("worker_id")
        session_id = request.data.get("session_id")
        notebook_id = request.data.get("notebook_id")

        filename = request.data.get("filename", "").replace(" ", "-")
        filepath = request.data.get("filepath")
        output_dir = request.data.get("output_dir")
        local_filepath = request.data.get("local_filepath")

        workers = Worker.objects.filter(
            pk=worker_id, session_id=session_id, notebook__id=notebook_id
        )
        if not workers:
            return Response(status=status.HTTP_404_NOT_FOUND)

        # remove previous objects with the same filepath
        WorkerFile.objects.filter(filepath=filepath, output_dir=output_dir).delete()

        # create a new object
        WorkerFile.objects.create(
            filename=filename,
            filepath=filepath,
            output_dir=output_dir,
            local_filepath=local_filepath,
            created_by=workers[0],
        )

        return Response(status=status.HTTP_200_OK)


class WorkerGetUploadedFilesUrls(APIView):
    def get(
        self,
        request,
        session_id,
        worker_id,
        notebook_id,
        format=None,
    ):
        try:
            # check if such worker exists
            worker = Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            client_action = "get_object"

            s3 = S3()
            files = UploadedFile.objects.filter(hosted_on=worker.notebook.hosted_on)
            urls = []
            for f in files:
                urls += [s3.get_presigned_url(f.filepath, client_action)]

            return Response({"urls": urls})

        except Exception as e:
            log.exception("Cant get uploaded files urls for worker")

        return Response(status=status.HTTP_403_FORBIDDEN)



import json

import ipywidgets
from IPython.display import display

from .manager import WidgetException, WidgetsManager


class Text:
    def __init__(
        self, value="", label="", rows=1, url_key="", disabled=False, hidden=False
    ):
        self.rows = rows

        self.code_uid = WidgetsManager.get_code_uid("Text", key=url_key)
        self.url_key = url_key
        self.hidden = hidden
        if WidgetsManager.widget_exists(self.code_uid):
            self.text = WidgetsManager.get_widget(self.code_uid)
            self.text.description = label
            self.text.disabled = disabled
        else:
            self.text = ipywidgets.Textarea(
                value=value, description=label, disabled=disabled
            )
            WidgetsManager.add_widget(self.text.model_id, self.code_uid, self.text)
        display(self)

    @property
    def value(self):
        return self.text.value

    # @value.setter
    # def value(self, v):
    #    self.text.value = v

    def __str__(self):
        return "mercury.Text"

    def __repr__(self):
        return "mercury.Text"

    def _repr_mimebundle_(self, **kwargs):
        # data = {}
        # data["text/plain"] = repr(self)
        # return data
        data = self.text._repr_mimebundle_()

        if len(data) > 1:
            view = {
                "widget": "Text",
                "value": self.text.value,
                "rows": self.rows,
                "label": self.text.description,
                "model_id": self.text.model_id,
                "code_uid": self.code_uid,
                "url_key": self.url_key,
                "disabled": self.text.disabled,
                "hidden": self.hidden,
            }
            data["application/mercury+json"] = json.dumps(view, indent=4)
            if "text/plain" in data:
                del data["text/plain"]

            if self.hidden:
                key = "application/vnd.jupyter.widget-view+json"
                if key in data:
                    del data[key]

            return data



from django.contrib import admin

from apps.accounts.models import Membership, Site


class SiteModelAdmin(admin.ModelAdmin):
    list_display = ("id", "title", "slug", "created_by")

    class Meta:
        model = Site


admin.site.register(Site, SiteModelAdmin)


class MembershipModelAdmin(admin.ModelAdmin):
    list_display = ("id", "user", "host", "rights")

    class Meta:
        model = Site


admin.site.register(Membership, MembershipModelAdmin)



# Please run tests with below command
# python manage.py test apps.nb.tests.test_nbrun
import os

from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from execnb.nbio import dict2nb
from apps.nb.nbrun import NbRun
from apps.nb.utils import one_cell_notebook


class NbRunTestCase(APITestCase):
    def setUp(self):
        pass

    def test_iframe_in_output(self):
        fname = "test-iframe-output.html"
        with open(fname, "w") as fout:
            fout.write("test")

        nbrun = NbRun()

        code = f"""from IPython.display import IFrame
IFrame(src="{fname}", width=100, height=100)"""

        nb = one_cell_notebook(code)
        nb = dict2nb(nb)

        nbrun.run_cell(nb.cells[0])
        self.assertTrue(fname not in nbrun.export_html(nb, full_header=False))

        os.remove(fname)



import json

import ipywidgets
from IPython.display import display

from .manager import WidgetsManager


class Select:
    def __init__(
        self, value=None, choices=[], label="", url_key="", disabled=False, hidden=False
    ):
        if value is None and len(choices) > 1:
            value = choices[0]

        self.code_uid = WidgetsManager.get_code_uid("Select", key=url_key)
        self.url_key = url_key
        self.hidden = hidden
        choices = list(choices)
        if WidgetsManager.widget_exists(self.code_uid):
            self.dropdown = WidgetsManager.get_widget(self.code_uid)
            if list(self.dropdown.options) != choices:
                self.dropdown.options = choices
                self.dropdown.value = value
            self.dropdown.description = label
            self.dropdown.disabled = disabled
        else:
            self.dropdown = ipywidgets.Dropdown(
                value=value,
                options=choices,
                description=label,
                style={"description_width": "initial"},
                disabled=disabled,
            )
            WidgetsManager.add_widget(
                self.dropdown.model_id, self.code_uid, self.dropdown
            )
        display(self)

    @property
    def value(self):
        return self.dropdown.value

    def __str__(self):
        return "m.Select"

    def __repr__(self):
        return "mercury.Select"

    def _repr_mimebundle_(self, **kwargs):
        # data = {}
        # data["text/plain"] = repr(self)
        # return data
        data = self.dropdown._repr_mimebundle_()

        if len(data) > 1:
            view = {
                "widget": "Select",
                "value": self.dropdown.value,
                "choices": self.dropdown.options,
                "label": self.dropdown.description,
                "model_id": self.dropdown.model_id,
                "code_uid": self.code_uid,
                "url_key": self.url_key,
                "disabled": self.dropdown.disabled,
                "hidden": self.hidden,
            }
            data["application/mercury+json"] = json.dumps(view, indent=4)
            if "text/plain" in data:
                del data["text/plain"]

            if self.hidden:
                key = "application/vnd.jupyter.widget-view+json"
                if key in data:
                    del data[key]

            return data



from django.contrib.auth.models import User
from django.db import models

from apps.accounts.fields import AutoCreatedField, AutoLastModifiedField
from apps.accounts.models import Site
from apps.workers.models import Worker
from apps.notebooks.models import Notebook


class UploadedFile(models.Model):
    """Files that are uploaded in Dashboard"""

    filename = models.CharField(max_length=1024, blank=False, null=False)
    filepath = models.CharField(max_length=1024, blank=False, null=False)
    filetype = models.CharField(max_length=128, blank=False, null=False)
    filesize = models.IntegerField(blank=False, null=False)  # size in B
    hosted_on = models.ForeignKey(Site, on_delete=models.CASCADE)
    created_at = AutoCreatedField()
    created_by = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
    )


class WorkerFile(models.Model):
    """Files created in worker (saved in output dir)"""

    filename = models.CharField(max_length=1024, blank=False, null=False)
    filepath = models.CharField(max_length=1024, blank=False, null=False)
    output_dir = models.CharField(max_length=1024, blank=False, null=False)
    local_filepath = models.CharField(max_length=1024, blank=False, null=False)
    created_at = AutoCreatedField()
    created_by = models.ForeignKey(
        Worker,
        on_delete=models.CASCADE,
    )


class UserUploadedFile(models.Model):
    """Files that are uploaded in notebook by users"""

    filename = models.CharField(max_length=1024, blank=False, null=False)
    filepath = models.CharField(max_length=1024, blank=False, null=False)

    hosted_on = models.ForeignKey(Site, on_delete=models.CASCADE)
    # web browser session id
    session_id = models.CharField(max_length=128)

    created_at = AutoCreatedField()



from django.test import TestCase

from apps.tasks.tasks import sanitize_string

# python manage.py test apps.tasks.tests -v 2


class SanitizeTestCase(TestCase):
    def test_CJK(self):
        input_string = """テスト(){}
                        asdfasdf"()"[]''{}:"""

        output_string = sanitize_string(input_string)

        self.assertTrue("(" not in output_string)
        self.assertTrue("[" not in output_string)
        self.assertTrue(output_string.startswith("テスト"))
        self.assertTrue(output_string.endswith("asdfasdf:"))



from dj_rest_auth.serializers import UserDetailsSerializer
from rest_framework import serializers

from apps.accounts.models import Invitation, Membership, Secret, Site, UserProfile


class UserProfileSerializer(serializers.ModelSerializer):
    class Meta:
        model = UserProfile
        fields = ("info",)


class UserSerializer(UserDetailsSerializer):
    profile = UserProfileSerializer()

    class Meta(UserDetailsSerializer.Meta):
        fields = UserDetailsSerializer.Meta.fields + ("profile",)


class SiteSerializer(serializers.ModelSerializer):
    created_by = UserDetailsSerializer(many=False, read_only=True)

    class Meta:
        model = Site
        read_only_fields = (
            "id",
            "created_at",
            "created_by",
            "updated_at",
            "status",
        )
        fields = (
            "id",
            "created_at",
            "created_by",
            "updated_at",
            "title",
            "slug",
            "domain",
            "custom_domain",
            "share",
            "welcome",
            "active",
            "status",
            "info",
            "domain",
        )


class MembershipSerializer(serializers.ModelSerializer):
    user = UserDetailsSerializer(many=False, read_only=True)
    created_by = UserDetailsSerializer(many=False, read_only=True)

    class Meta:
        model = Membership
        read_only_fields = ("id", "created_at", "created_by", "updated_at")
        fields = read_only_fields + ("rights", "user")


class InvitationSerializer(serializers.ModelSerializer):
    created_by = UserDetailsSerializer(many=False, read_only=True)

    class Meta:
        model = Invitation
        read_only_fields = ("id", "invited", "created_at", "created_by", "rights")
        fields = read_only_fields


class SecretSerializer(serializers.ModelSerializer):
    created_by = UserDetailsSerializer(many=False, read_only=True)

    class Meta:
        model = Secret
        read_only_fields = ("id", "name", "created_at", "created_by")
        fields = read_only_fields



import asyncio
import concurrent.futures
import logging
import os
import tempfile
from subprocess import PIPE, Popen

from pyppeteer import launch


async def html_to_png(html_file, png_file, pyppeteer_args=None):
    """Convert a HTML file to a PDF"""
    browser = await launch(
        handleSIGINT=False,
        handleSIGTERM=False,
        handleSIGHUP=False,
        headless=True,
        args=["--no-sandbox"],
    )

    page = await browser.newPage()
    await page.setViewport(dict(width=600, height=400))
    await page.emulateMedia("screen")

    await page.goto(f"file://{html_file}", {"waitUntil": ["networkidle2"]})

    page_margins = {
        "left": "20px",
        "right": "20px",
        "top": "30px",
        "bottom": "30px",
    }

    dimensions = await page.evaluate(
        """() => {
        return {
            width: document.body.scrollWidth,
            height: document.body.scrollHeight,
            offsetWidth: document.body.offsetWidth,
            offsetHeight: document.body.offsetHeight,
            deviceScaleFactor: window.devicePixelRatio,
        }
    }"""
    )
    width = dimensions["width"]
    height = dimensions["height"]

    await page.evaluate(
        """
    function getOffset( el ) {
        var _x = 0;
        var _y = 0;
        while( el && !isNaN( el.offsetLeft ) && !isNaN( el.offsetTop ) ) {
            _x += el.offsetLeft - el.scrollLeft;
            _y += el.offsetTop - el.scrollTop;
            el = el.offsetParent;
        }
        return { top: _y, left: _x };
        }
    """,
        force_expr=True,
    )

    await page.addStyleTag(
        {
            "content": """
                #notebook-container {
                    box-shadow: none;
                    padding: unset
                }
                div.cell {
                    page-break-inside: avoid;
                    break-inside: avoid;
                }
                div.output_wrapper {
                    page-break-inside: avoid;
                    break-inside: avoid;
                }
                div.output {
                    page-break-inside: avoid;
                    break-inside: avoid;
                }
                /* Jupyterlab based HTML uses these classes */
                .jp-Cell-inputWrapper {
                    page-break-inside: avoid;
                    break-inside: avoid;
                }
                .jp-Cell-outputWrapper {
                    page-break-inside: avoid;
                    break-inside: avoid;
                }
                .jp-Notebook {
                    margin: 0px;
                }
                /* Hide the message box used by MathJax */
                #MathJax_Message {
                    display: none;
                }
         """
        }
    )

    await page.screenshot(
        {
            "path": png_file,
            # "clip": {
            #     "x": 0,
            #     "y": 0,
            #     "width": 500,
            #     "height": 500
            # }
        }
    )

    await browser.close()


def install_chromium():
    command = ["pyppeteer-install"]
    with Popen(command, stdout=PIPE, stderr=PIPE) as proc:
        print(proc.stdout.read())
        print(proc.stderr.read())


def to_png(html_input_file, png_output_file):
    # make sure chromium is installed
    # install_chromium()

    # dont want to see DEBUG logs for chromium ...
    prev_log_level = logging.getLogger().getEffectiveLevel()
    logging.getLogger().setLevel(logging.ERROR)

    # convert notebook to PDF
    pool = concurrent.futures.ThreadPoolExecutor()
    pool.submit(
        asyncio.run,
        html_to_png(html_input_file, png_output_file),
    ).result()

    # set previous log level
    logging.getLogger().setLevel(prev_log_level)



import json

import ipywidgets
from IPython.display import display

from .manager import WidgetException, WidgetsManager


class Button:
    def __init__(self, label="", style="primary", disabled=False, hidden=False):
        self.code_uid = WidgetsManager.get_code_uid("Button")

        if style not in ["primary", "success", "info", "warning", "danger", ""]:
            style = "primary"
        self.hidden = hidden

        if WidgetsManager.widget_exists(self.code_uid):
            self.button = WidgetsManager.get_widget(self.code_uid)
            self.button.description = label
            self.button.button_style = style
            self.button.disabled = disabled
        else:
            self.button = ipywidgets.Button(
                description=label, button_style=style, disabled=disabled
            )
            self.button.value = False

            def on_button_clicked(b):
                self.button.value = True

            self.button.on_click(on_button_clicked)

            WidgetsManager.add_widget(self.button.model_id, self.code_uid, self.button)

        display(self)

    @property
    def clicked(self):
        if self.button.value:
            self.button.value = False
            return True
        return False

    def __str__(self):
        return "mercury.Button"

    def __repr__(self):
        return "mercury.Button"

    def _repr_mimebundle_(self, **kwargs):
        # data = {}
        # data["text/plain"] = repr(self)
        # return data
        data = self.button._repr_mimebundle_()

        if len(data) > 1:
            view = {
                "widget": "Button",
                "label": self.button.description,
                "style": self.button.button_style,
                "value": False,
                "model_id": self.button.model_id,
                "code_uid": self.code_uid,
                "disabled": self.button.disabled,
                "hidden": self.hidden,
            }
            data["application/mercury+json"] = json.dumps(view, indent=4)
            if "text/plain" in data:
                del data["text/plain"]

            if self.hidden:
                key = "application/vnd.jupyter.widget-view+json"
                if key in data:
                    del data[key]

            return data



import logging

from django.db.models import Q
from django.http import JsonResponse
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Membership, Site
from apps.notebooks.models import Notebook
from apps.storage.models import UserUploadedFile
from apps.storage.s3utils import S3
from apps.storage.serializers import UploadedFileSerializer
from apps.workers.models import Worker
from apps.accounts.views.utils import is_cloud_version

from apps.accounts.views.sites import (
    get_plan,
    PLAN_KEY,
    PLAN_STARTER,
    PLAN_PRO,
    PLAN_BUSINESS,
)

from apps.storage.utils import get_bucket_key, get_user_upload_bucket_key
from apps.storage.views.dashboardfiles import FILE_LIMITS

log = logging.getLogger(__name__)


def user_upload_allowed(user, site_id):
    try:
        sites = Site.objects.filter(pk=site_id)

        if not sites:
            return False
        if sites[0].share == Site.PUBLIC:
            return True
        if sites[0].share == Site.PRIVATE and user.is_anonymous:
            return False

        sites = sites.filter(
            Q(pk__in=Membership.objects.filter(user=user).values("host__id"))
            | Q(created_by=user)
        )
        if not sites:
            return False
        return True
    except Exception as e:
        pass
    return False


def upload_allowed_check_limits(user, site_id, filesize):
    try:
        if not is_cloud_version():
            return True

        site = Site.objects.get(pk=site_id)
        plan = get_plan(site.created_by)
        files_size_limit = FILE_LIMITS[plan]["size"]  # in MB

        if int(filesize) / 1024 / 1024 > files_size_limit:
            return False
        return True
    except Exception as e:
        pass
    return False


class NbPresignedUrlPut(APIView):
    def get(self, request, site_id, session_id, filename, filesize, format=None):
        site_id = int(site_id)

        if not user_upload_allowed(request.user, site_id):
            return Response(status=status.HTTP_403_FORBIDDEN)

        if not upload_allowed_check_limits(request.user, site_id, filesize):
            return Response(status=status.HTTP_403_FORBIDDEN)

        client_action = "put_object"
        s3 = S3()
        url = s3.get_presigned_url(
            get_user_upload_bucket_key(site_id, session_id, filename.replace(" ", "-")),
            client_action,
        )
        return Response({"url": url})


class NbFileUploaded(APIView):
    def post(self, request, format=None):
        site_id = int(request.data.get("site_id"))
        session_id = request.data.get("session_id")
        filename = request.data.get("filename", "").replace(" ", "-")

        bucket_key = get_user_upload_bucket_key(site_id, session_id, filename)

        s3 = S3()
        if not s3.file_exists(bucket_key):
            return Response(status=status.HTTP_403_FORBIDDEN)

        site = Site.objects.get(pk=site_id)
        # remove previous objects with the same filepath
        UserUploadedFile.objects.filter(
            filepath=bucket_key, session_id=session_id, hosted_on=site
        ).delete()

        # create a new object
        upload = UserUploadedFile.objects.create(
            filename=filename,
            filepath=bucket_key,
            session_id=session_id,
            hosted_on=site,
        )

        return Response(status=status.HTTP_200_OK)


class NbDeleteFile(APIView):
    def post(self, request, format=None):
        site_id = int(request.data.get("site_id"))
        session_id = request.data.get("session_id")
        filename = request.data.get("filename", "").replace(" ", "-")
        bucket_key = get_user_upload_bucket_key(site_id, session_id, filename)

        s3 = S3()
        s3.delete_file(bucket_key)

        UserUploadedFile.objects.filter(
            filepath=bucket_key, session_id=session_id, hosted_on__id=site_id
        ).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


class WorkerGetNbFileUrl(APIView):
    def get(
        self,
        request,
        session_id,
        worker_id,
        notebook_id,
        filename,
        format=None,
    ):
        try:
            # check if such worker exists
            worker = Worker.objects.get(
                pk=int(worker_id), session_id=session_id, notebook__id=int(notebook_id)
            )

            upload = UserUploadedFile.objects.filter(
                filename=filename.replace(" ", "-"), session_id=session_id
            ).latest("id")

            client_action = "get_object"
            s3 = S3()

            return Response(
                {"url": s3.get_presigned_url(upload.filepath, client_action)}
            )

        except Exception as e:
            log.exception("Cant get user uploaded file url for worker")

        return Response(status=status.HTTP_403_FORBIDDEN)



import copy
import hashlib
import json
import logging
import os
import sys
import threading
import time
from datetime import datetime

from execnb.nbio import nb2dict, read_nb

from apps.nb.nbrun import NbRun
from apps.nbworker.utils import Purpose, stop_event
from apps.nbworker.ws import WSClient
from apps.workers.models import WorkerState
from apps.ws.utils import parse_params
from widgets.manager import WidgetsManager

log = logging.getLogger(__name__)


class NBWorker(WSClient):
    def __init__(self, ws_address, notebook_id, session_id, worker_id):
        super(NBWorker, self).__init__(ws_address, notebook_id, session_id, worker_id)

        self.prev_nb = None
        self.prev_widgets = {}
        self.prev_body = ""
        self.prev_update_time = None
        self.prev_md5 = None
        self.start_time = time.time()
        self.max_idle_time = int(
            os.environ.get("MAX_IDLE_TIME", 12 * 60 * 60)
        )  # default idle after 12 hours
        self.max_run_time = int(
            os.environ.get("MAX_RUN_TIME", 24 * 60 * 60)
        )  # default max run time is 24 hours
        self.last_execution_time = time.time()
        self.start_time = time.time()

        # monitor notebook file updates if running locally
        if "127.0.0.1" in ws_address:
            threading.Thread(target=self.nb_file_watch, daemon=True).start()
        threading.Thread(target=self.process_msgs, daemon=True).start()
        self.ws.run_forever(ping_interval=10, ping_timeout=5)

    @staticmethod
    def md5(fname):
        hash_md5 = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def nb_file_watch(self):
        while not stop_event.is_set():
            current_update_time = datetime.fromtimestamp(
                os.path.getmtime(self.notebook.path)
            )
            if (
                self.prev_update_time is not None
                and self.prev_update_time != current_update_time
            ):
                checksum = NBWorker.md5(self.notebook.path)

                log.debug(f"Checksum {checksum} prev {self.prev_md5}")

                if self.prev_md5 is None or checksum != self.prev_md5:
                    log.debug("Notebook file changed!")
                    msg = json.dumps({"purpose": Purpose.InitNotebook})
                    self.queue.put(msg)

                self.prev_md5 = checksum

            self.prev_update_time = current_update_time
            time.sleep(0.25)

    def process_msgs(self):
        global stop_event
        while not stop_event.is_set():
            item = self.queue.get()
            log.debug(f"Porcess msg {item}")
            json_data = json.loads(item)

            self.last_execution_time = time.time()

            if json_data.get("purpose", "") == Purpose.InitNotebook:
                self.init_notebook()
            elif json_data.get("purpose", "") == Purpose.RunNotebook:
                self.run_notebook(json_data)
            elif json_data.get("purpose", "") == Purpose.DisplayNotebook:
                self.display_notebook(json_data)
            elif json_data.get("purpose", "") == Purpose.ClearSession:
                self.init_notebook()
            elif json_data.get("purpose", "") == Purpose.CloseWorker:
                stop_event.set()
                self.delete_worker()
                sys.exit(1)
            elif json_data.get("purpose", "") == Purpose.DownloadHTML:
                self.download_html()
            elif json_data.get("purpose", "") == Purpose.DownloadPDF:
                self.download_pdf()

            self.queue.task_done()

    def worker_pong(self):
        total_run_time = time.time() - self.start_time
        elapsed_from_last_execution = time.time() - self.last_execution_time
        log.debug(f"Total run time {total_run_time}")
        log.debug(f"Elapsed from last execution {elapsed_from_last_execution}")

        close_worker = True
        if total_run_time > self.max_run_time:
            self.update_worker_state(WorkerState.MaxRunTimeReached)
        elif elapsed_from_last_execution > self.max_idle_time:
            self.update_worker_state(WorkerState.MaxIdleTimeReached)
        else:
            self.update_worker_state(self.worker_state())
            close_worker = False

        if self.worker_exists():
            self.send_state()

        if close_worker:
            self.queue.put(json.dumps({"purpose": Purpose.CloseWorker}))

    def run_notebook(self, json_params):
        log.debug(f"Run notebook with {json_params}")
        self.update_worker_state(WorkerState.Busy)

        widgets = json.loads(json_params.get("widgets", "{}"))

        self.update_nb(widgets)

        self.sm.sync_output_dir()

        body = self.nbrun.export_html(self.nb, full_header=self.is_presentation())

        # with open(f"test_{counter}.html", "w") as fout:
        #    fout.write(body)

        self.ws.send(json.dumps({"purpose": Purpose.ExecutedNotebook, "body": body}))
        self.update_worker_state(WorkerState.Running)
        self.prev_body = copy.deepcopy(body)

    def update_nb(self, widgets):
        log.debug(f"Update nb {widgets}")

        index_execute_from = None
        # fill notebook with widgets values
        self.prev_widgets = copy.deepcopy(widgets)
        for widget_key in widgets.keys():
            value = widgets[widget_key]

            widget_type = WidgetsManager.parse_widget_type(widget_key)
            log.debug(
                f"Update widget code_uid={widget_key} value={value} widget type {widget_type}"
            )

            if widget_type == "File" and len(value) == 2:
                log.debug(f"Get file {value[0]} from id={value[1]}")
                # tu = TemporaryUpload.objects.get(upload_id=value[1])
                # value[1] = tu.get_file_path()
                # value[1] = value[1].replace("\\", "\\\\")
                value = self.sm.get_user_uploaded_file(value)
                log.debug(f"File path is {value[1]}")

                code = (
                    f'WidgetsManager.update("{widget_key}", field="filename", new_value="{value[0]}")\n'
                    f'WidgetsManager.update("{widget_key}", field="filepath", new_value="{value[1]}")\n'
                )
            elif widget_type == "OutputDir":
                output_dir = self.sm.worker_output_dir()
                output_dir = output_dir.replace("\\", "\\\\")
                code = f'WidgetsManager.update("{widget_key}", field="value", new_value="{output_dir}")'

            elif isinstance(value, str):
                code = f'WidgetsManager.update("{widget_key}", field="value", new_value="""{value}""")'
            else:
                code = f'WidgetsManager.update("{widget_key}", field="value", new_value={value})'

            log.debug(f"Execute code {code}")

            r = self.nbrun.run_code(code)

            updated = "True" in str(r)
            log.debug(f"Update reponse {r}, updated={updated}")

            if updated:
                cell_index = WidgetsManager.parse_cell_index(widget_key)
                log.debug(f"Widget updated, update nb from {cell_index}")

                if index_execute_from is None:
                    index_execute_from = cell_index
                else:
                    index_execute_from = min(index_execute_from, cell_index)

        if index_execute_from is not None:
            if self.prev_nb is not None:
                self.nb = copy.deepcopy(self.prev_nb)
            else:
                self.nb = copy.deepcopy(self.nb_original)

            self.nbrun.run_notebook(self.nb, start=index_execute_from - 1)

            self.send_widgets(self.nb, expected_widgets_keys=widgets.keys())
            self.prev_nb = copy.deepcopy(self.nb)
        else:
            log.debug("Skip nb execution, no changes in widgets")

    def send_widgets(self, nb, expected_widgets_keys, init_widgets=False):
        nb_widgets_keys = []
        widgets_params = []
        for cell in nb.cells:
            for output in cell.get("outputs", []):
                if "data" in output:
                    if "application/mercury+json" in output["data"]:
                        w = output["data"]["application/mercury+json"]
                        log.debug(w)
                        w = json.loads(w)

                        # prepare msg to send by ws
                        msg = WidgetsManager.frontend_format(w)
                        if msg:
                            msg["widgetKey"] = w.get("code_uid")
                            log.debug(f"Update widget {msg}")
                            if init_widgets:
                                widgets_params += [msg]
                            else:
                                msg["purpose"] = Purpose.UpdateWidgets
                                self.ws.send(json.dumps(msg))

                        code_uid = w.get("code_uid")
                        if code_uid is not None:
                            nb_widgets_keys += [code_uid]

                        if w.get("widget", "") == "App":
                            if w.get("title", "") != "":
                                self.ws.send(
                                    json.dumps(
                                        {
                                            "purpose": Purpose.UpdateTitle,
                                            "title": w.get("title"),
                                        }
                                    )
                                )
                            if w.get("show_code") is not None:
                                self.nbrun.set_show_code(w.get("show_code"))
                                self.ws.send(
                                    json.dumps(
                                        {
                                            "purpose": Purpose.UpdateShowCode,
                                            "showCode": w.get("show_code"),
                                        }
                                    )
                                )

        if init_widgets:
            msg = {"purpose": Purpose.InitWidgets, "widgets": widgets_params}
            log.debug("Init widgets")
            log.debug(msg)
            self.ws.send(json.dumps(msg))
        else:
            # check if hide some widgets
            # needed only when updating widgets
            hide_widgets = []
            for widget_key in expected_widgets_keys:
                if widget_key not in nb_widgets_keys:
                    hide_widgets += [widget_key]

            log.debug(f"Hide widgets {hide_widgets}")
            if hide_widgets:
                msg = {"purpose": Purpose.HideWidgets, "keys": hide_widgets}
                self.ws.send(json.dumps(msg))

    def initialize_outputdir(self):
        output_dir = self.sm.worker_output_dir()
        output_dir = output_dir.replace("\\", "\\\\")
        self.nbrun.run_code(
            f"""import os\nos.environ["MERCURY_OUTPUTDIR"]="{output_dir}" """
        )

    def install_new_packages(self):
        if "127.0.0.1" not in self.ws_address and "localhost" not in self.ws_address:
            fname = "requirements.txt"
            if os.path.exists(fname):
                log.debug(f"Install new packages from requirements.txt")
                cmd = f"pip install -r {fname}"
                self.nbrun.run_code(cmd)

    def provision_secrets(self):
        secrets = self.list_secrets()
        if secrets:
            cmd = "import os\n"
            for s in secrets:
                name = s.get("name", "")
                secret = s.get("secret", "")
                cmd += f'os.environ["{name}"] = "{secret}"'
            log.debug("Set secrets")
            self.nbrun.run_code(cmd)

    def init_notebook(self):
        log.debug(f"Init notebook, show_code={self.show_code()}")

        self.sm.provision_uploaded_files()

        self.prev_nb = None
        self.prev_widgets = {}
        self.prev_body = ""

        self.update_worker_state(WorkerState.Busy)

        self.nbrun = NbRun(
            show_code=self.show_code(),
            show_prompt=self.show_prompt(),
            is_presentation=self.is_presentation(),
            reveal_theme=self.reveal_theme(),
        )

        self.install_new_packages()
        self.provision_secrets()

        # we need to initialize the output dir always
        # even if there is no OutputDir in the notebook
        self.initialize_outputdir()

        self.nb_original = read_nb(self.notebook.path)

        self.nbrun.run_notebook(self.nb_original)

        self.sm.sync_output_dir()

        # TODO: update params in db if needed"
        params = {}
        parse_params(nb2dict(self.nb_original), params)

        # update database ...

        log.debug(f"Executed params {json.dumps(params, indent=4)}")
        update_database = self.update_notebook(params)

        # update_database = False
        # if params.get("title", "") != "" and self.notebook.title != params.get(
        #     "title", ""
        # ):
        #     self.notebook.title = params.get("title", "")
        #     update_database = True

        # nb_params = json.loads(self.notebook.params)
        # for property in [
        #     "show-code",
        #     "show-prompt",
        #     "continuous_update",
        #     "static_notebook",
        #     "description",
        #     "show_sidebar",
        #     "full_screen",
        #     "allow_download",
        # ]:
        #     if params.get(property) is not None and nb_params.get(
        #         property
        #     ) != params.get(property):
        #         nb_params[property] = params.get(property)
        #         update_database = True
        # # save widgets params
        # if json.dumps(nb_params.get("params", {})) != json.dumps(
        #     params.get("params", {})
        # ):
        #     nb_params["params"] = params["params"]
        #     update_database = True

        # if update_database:
        #     self.notebook.params = json.dumps(nb_params)
        #     self.notebook.save()

        nb_params = json.loads(self.notebook.params)

        self.nbrun.set_show_code_and_prompt(
            nb_params.get("show-code", False), nb_params.get("show-prompt", True)
        )
        self.nbrun.set_is_presentation(nb_params.get("output", "app") == "slides")

        log.debug(params)
        log.debug(f"Exporter show_code {self.nbrun.exporter.show_code}")

        self.nb = copy.deepcopy(self.nb_original)

        if self.is_presentation():
            body = self.nbrun.export_html(self.nb, full_header=True)
        else:
            body = self.nbrun.export_html(self.nb, full_header=False)

        msg = {"purpose": Purpose.ExecutedNotebook, "body": body}
        if update_database:
            msg["reloadNotebook"] = True
        self.ws.send(json.dumps(msg))
        self.prev_body = copy.deepcopy(body)

        self.send_widgets(self.nb, expected_widgets_keys=[], init_widgets=True)
        self.update_worker_state(WorkerState.Running)

    # def save_notebook(self):
    #     log.debug(f"Save notebook")
    #     # save nb in HTML
    #     if self.is_presentation():
    #         nb_body = self.nbrun.export_html(self.nb, full_header=True)
    #     else:
    #         nb_body = self.nbrun.export_html(self.nb, full_header=True)

    #     sm = StorageManager(self.session_id, self.worker_id)
    #     nb_path = sm.save_nb_html(nb_body)

    #     # create task with path to HTML file
    #     task = Task.objects.create(
    #         task_id=f"worker-{self.worker_id}",
    #         session_id=self.session_id,
    #         notebook_id=self.notebook_id,
    #         state="DONE",
    #         params=json.dumps(self.prev_widgets),
    #         result=nb_path,
    #     )
    #     log.debug(f"Task ({task.id}) created")

    #     # send notice that nb saved
    #     self.ws.send(json.dumps({"purpose": Purpose.SavedNotebook}))

    def display_notebook(self, json_params):
        log.debug(f"Display notebook ({json_params})")

    def download_html(self):
        log.debug(f"Download HTML")
        # save nb in HTML with full header
        if self.is_presentation():
            nb_body = self.nbrun.export_html(self.nb, full_header=True)
        else:
            nb_body = self.nbrun.export_html(self.nb, full_header=True)

        _, nb_url = self.sm.save_nb_html(nb_body)

        # send HTML url address
        self.ws.send(
            json.dumps(
                {
                    "purpose": Purpose.DownloadHTML,
                    "url": nb_url,
                    "filename": f"{self.notebook.slug}.html",
                }
            )
        )

    def download_pdf(self):
        log.debug(f"Download PDF")
        # save nb in HTML with full header
        nb_body = self.nbrun.export_html(self.nb, full_header=True)

        # export to PDF
        _, pdf_url = self.sm.save_nb_pdf(nb_body, self.is_presentation())

        # send PDF url
        self.ws.send(
            json.dumps(
                {
                    "purpose": Purpose.DownloadPDF,
                    "url": pdf_url,
                    "filename": f"{self.notebook.slug}.pdf",
                }
            )
        )



import os
import sys
import uuid

from celery import Celery
from celery.schedules import crontab
from django.conf import settings
from django.db import transaction

CURRENT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, CURRENT_DIR)

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "server.settings")

#  celery -A server worker --loglevel=info -P gevent --concurrency 1 -E
app = Celery("server")

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("django.conf:settings", namespace="CELERY")

app.conf.timezone = settings.TIME_ZONE

app.autodiscover_tasks()

# workers with active websocket connection are in the separate queue
app.conf.task_routes = {
    "apps.ws.tasks.task_start_websocket_worker": {"queue": "ws"}
}


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    try:
        import django

        django.setup()
        from apps.notebooks.models import Notebook

        # get all notebooks with not empty schedule
        notebooks = Notebook.objects.exclude(schedule__isnull=True).exclude(
            schedule__exact=""
        )

        for n in notebooks:
            schedule_str = n.schedule
            schedule_arr = schedule_str.split(" ")
            minute, hour, day_of_month, month, day_of_week = (
                schedule_arr[0],
                schedule_arr[1],
                schedule_arr[2],
                schedule_arr[3],
                schedule_arr[4],
            )
            sender.add_periodic_task(
                crontab(
                    minute=minute,
                    hour=hour,
                    day_of_month=day_of_month,
                    month_of_year=month,
                    day_of_week=day_of_week,
                ),
                execute_notebook.s(n.id),
            )
    except Exception as e:
        print("Problem with periodic tasks setup")
        print(str(e))


@app.task
def execute_notebook(notebook_id):
    import django

    django.setup()
    from apps.notebooks.models import Notebook
    from apps.tasks.models import Task
    from apps.tasks.tasks import task_execute

    with transaction.atomic():
        task = Task(
            session_id=f"scheduled-{uuid.uuid4().hex[:8]}",
            state="CREATED",
            notebook=Notebook.objects.get(pk=notebook_id),
            params="{}",
        )
        task.save()
        job_params = {"db_id": task.id}
        transaction.on_commit(lambda: task_execute.delay(job_params))



from IPython.display import display, HTML


def Confetti():
    display(
        HTML(
            """
    <script type="module">import confetti from 'https://cdn.skypack.dev/canvas-confetti'; 
    confetti({
    particleCount: 300, 
    angle: 90, 
    spread: 180,
    decay: 0.9,
    startVelocity: 40,
    origin: {
        x: 0.5,
        y: 0.5
    }
    });
    
    </script>
    """
        )
    )



import os


class StopExecution(Exception):
    def _render_traceback_(self):
        if os.environ.get("RUN_MERCURY") is not None:
            return ["StopExecution"]
        pass


def Stop():
    raise StopExecution()



import json
import logging

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer

from apps.workers.models import Worker
from apps.ws.utils import client_group, worker_group

log = logging.getLogger(__name__)


class WorkerProxy(WebsocketConsumer):
    def connect(self):
        self.notebook_id = int(self.scope["url_route"]["kwargs"]["notebook_id"])
        self.session_id = self.scope["url_route"]["kwargs"]["session_id"]
        self.worker_id = self.scope["url_route"]["kwargs"]["worker_id"]

        # check if there is such worker requested in database
        workers = Worker.objects.filter(
            pk=self.worker_id, notebook__id=self.notebook_id, session_id=self.session_id
        )
        if not workers:
            self.close()

        log.debug(
            f"Worker ({self.worker_id}) connect to {self.session_id}, notebook id {self.notebook_id}"
        )

        self.client_group = client_group(self.notebook_id, self.session_id)
        self.worker_group = worker_group(self.notebook_id, self.session_id)

        async_to_sync(self.channel_layer.group_add)(
            self.worker_group, self.channel_name
        )

        self.accept()

    def disconnect(self, close_code):
        async_to_sync(self.channel_layer.group_discard)(
            self.worker_group, self.channel_name
        )

    def receive(self, text_data):
        json_data = json.loads(text_data)
        # broadcast to all clients
        async_to_sync(self.channel_layer.group_send)(
            self.client_group, {"type": "broadcast_message", "payload": json_data}
        )

    def broadcast_message(self, event):
        payload = event["payload"]
        self.send(text_data=json.dumps(payload))



import os
from uuid import uuid4


class NumberBox:
    BLUE = "#00B1E4"
    LIGHT_BLUE = "rgba(0, 177, 228, 0.5)"
    RED = "#FF6384"
    LIGHT_RED = "rgba(255, 99, 132, 0.5)"
    GREEN = "#00B275"
    LIGHT_GREEN = "rgb(0, 178, 117, 0.5)"

    def __init__(
        self,
        data,
        title="",
        percent_change=None,
        background_color="white",
        border_color="lightgray",
        data_color="black",
        title_color="gray",
    ):
        self.data = data
        self.title = title
        self.blox_type = "numeric"
        if isinstance(self.data, list):
            self.blox_type = "list"
        self.percent_change = percent_change
        self.background_color = background_color
        self.border_color = border_color
        self.data_color = data_color
        self.title_color = title_color
        # position when displayed as a list
        self.position = None

    def styles(self):
        return """<style scoped>
        .numberbox-container {
        width: 100%;
        display: flex;
        flex-direction: row;
        }
        @media (max-width: 800px) {
        .numberbox-container {
            flex-direction: column;
        }
        }
        </style>"""

    def _repr_html_(self):
        if self.blox_type == "list":
            bloxs = ""
            for i, b in enumerate(self.data):
                if isinstance(b, NumberBox):
                    # we dont set position for last item
                    # because we dont need to add margin for it
                    if i != len(self.data)-1:
                        b.position = i
                    bloxs += b._repr_html_()

            return f"""{self.styles()}<div class="numberbox-container" style="display: flex; background: #fff;">{bloxs}</div>"""

        percent_change_html = ""
        if self.percent_change is not None:
            if self.percent_change > 0:
                percent_change_html = f"""
                <span style="font-size: 1.3em; color: {self.GREEN}; font-family: monospace;"> +{self.percent_change}%</span>
                """
            else:
                percent_change_html = f"""
                <span style="font-size: 1.3em; color: {self.RED}; font-family: monospace;"> {self.percent_change}%</span>
                """
        title_html = ""
        if self.title != "":
            title_html = f"""<span style="font-size: 2em; color: {self.title_color}; display: block; padding-top: 20px; font-family: monospace; line-height: 1.3em;">{self.title}</span>"""

        data_str = ""
        if isinstance(self.data, str):
            data_str = self.data
        else:
            data_str = f"{self.data:,}"

        margin = "0px" if self.position is None else "15px"
        return f"""
<div style="text-align: center; width: 100%; border: 1px solid {self.border_color}; margin-right: {margin}; margin-top: 15px; padding-top: 40px; padding-bottom: 30px; background: {self.background_color}; border-radius:5px">
  <span style="font-size: 4em; color: {self.data_color}; font-family: monospace; ">{data_str}</span>
  {percent_change_html}
  {title_html}
</div>
  """



import json

import ipywidgets
from IPython.display import Markdown, display

from .manager import WidgetException, WidgetsManager


class NoteText:
    def __init__(self, text):
        self.value = text


class Note:
    def __init__(self, text=""):
        self.code_uid = WidgetsManager.get_code_uid("Note")

        if WidgetsManager.widget_exists(self.code_uid):
            self.note = WidgetsManager.get_widget(self.code_uid)
            if self.note.value != text:
                self.note.value = text
        else:
            self.note = NoteText(text)
            WidgetsManager.add_widget(self.code_uid, self.code_uid, self.note)
        display(self)

    @property
    def value(self):
        return self.note.value

    def __str__(self):
        return "mercury.Note"

    def __repr__(self):
        return "mercury.Note"

    def _repr_mimebundle_(self, **kwargs):
        data = {}

        view = {
            "widget": "Note",
            "value": self.note.value,
            "model_id": self.code_uid,
            "code_uid": self.code_uid,
        }
        data["application/mercury+json"] = json.dumps(view, indent=4)

        data["text/markdown"] = self.note.value

        return data



import logging
from tempfile import TemporaryFile

import boto3
from django.conf import settings

log = logging.getLogger(__name__)


class S3:
    def __init__(self):
        try:
            self.s3_client = boto3.client(
                "s3",
                aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
                aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
                region_name=settings.AWS_REGION_NAME,
            )
        except Exception:
            log.exception("Cant create S3 client")

    def get_presigned_url(
        self, bucket_key, client_method="put_object", expires_in=6 * 3600
    ):
        try:
            method_parameters = {
                "Bucket": settings.AWS_BUCKET_NAME,
                "Key": bucket_key,
            }
            url = self.s3_client.generate_presigned_url(
                ClientMethod=client_method,
                Params=method_parameters,
                ExpiresIn=expires_in,
            )
            return url
        except Exception:
            log.exception(
                f"Exception when get presigned url, {bucket_key}, {client_method}"
            )
        return None

    def download_file(self, bucket_key, file_name):
        try:
            self.s3_client.download_file(
                settings.AWS_BUCKET_NAME, bucket_key, file_name
            )
            return True
        except Exception:
            log.exception(f"Exception when downloading file {bucket_key}")
        return False

    def upload_file(self, file_name, bucket_key):
        try:
            ExtraArgs = {}
            if file_name.endswith("html"):
                ExtraArgs = {"ContentType": "text/html"}
            self.s3_client.upload_file(
                file_name, settings.AWS_BUCKET_NAME, bucket_key, ExtraArgs=ExtraArgs
            )
            return True
        except Exception:
            log.exception(f"Exception when uploading file {file_name}")
        return False

    def delete_file(self, bucket_key):
        try:
            self.s3_client.delete_object(
                Bucket=settings.AWS_BUCKET_NAME, Key=bucket_key
            )
            return True
        except Exception as e:
            log.exception(f"Exception when delete file {bucket_key}")
        return False

    def list_files(self, prefix):
        try:
            files = []
            for key in self.s3_client.list_objects(
                Bucket=settings.AWS_BUCKET_NAME, Prefix=prefix
            ).get("Contents", []):
                files += [key["Key"]]
        except Exception:
            log.exception(f"Exception when list files from {prefix}")
        return files

    def file_exists(self, bucket_key):
        try:
            prefix = "/".join(bucket_key.split("/")[:-1])
            files = self.list_files(prefix)
            print(files, bucket_key, bucket_key in files)
            return bucket_key in files
        except Exception:
            log.exception(f"Exception when check if file exists {bucket_key}")
        return False


def clean_worker_files(site_id, session_id):
    if settings.STORAGE == settings.STORAGE_S3:
        s3 = S3()
        # remove output dir files
        bucket_key = f"session-{session_id}"
        keys = s3.list_files(bucket_key)
        for key in keys:
            s3.delete_file(key)
        # remove user upload files
        bucket_key = f"site-{site_id}/session-{session_id}/user-input"
        keys = s3.list_files(bucket_key)
        for key in keys:
            s3.delete_file(key)



# Please run tests with below command
# python manage.py test apps.accounts
import os
from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core import mail
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Membership, Secret, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker


# tests are disabled because they use live paddle api

# class SubscriptionTestCase(APITestCase):
#     register_url = "/api/v1/auth/register/"
#     verify_email_url = "/api/v1/auth/register/verify-email/"
#     login_url = "/api/v1/auth/login/"
#     user_details_url = "/api/v1/auth/user/"

#     subscription_url = "/api/v1/subscription"

#     def setUp(self):
#         #
#         # first user
#         #
#         self.user1_params = {
#             "username": "user1",  # it is optional to pass username
#             "email": "piotr@example.com",
#             "password": "verysecret",
#         }
#         self.user = User.objects.create_user(
#             username=self.user1_params["username"],
#             email=self.user1_params["email"],
#             password=self.user1_params["password"],
#         )
#         EmailAddress.objects.create(
#             user=self.user, email=self.user.email, verified=True, primary=True
#         )
#         #
#         # second user
#         #
#         self.user2_params = {
#             "username": "user2",  # it is optional to pass username
#             "email": "piotr2@example.com",
#             "password": "verysecret2",
#         }
#         self.user2 = User.objects.create_user(
#             username=self.user2_params["username"],
#             email=self.user2_params["email"],
#             password=self.user2_params["password"],
#         )
#         EmailAddress.objects.create(
#             user=self.user2, email=self.user2.email, verified=True, primary=True
#         )

#     def test_check_subscription(self):
#         # login as second user to get token
#         response = self.client.post(self.login_url, self.user1_params)
#         token = response.json()["key"]
#         headers = {"HTTP_AUTHORIZATION": "Token " + token}

#         new_data = {"action": "check", "checkoutId": "206702119-chre59f1e77e6ca-132bd39c55"}
#         response = self.client.post(self.subscription_url, new_data, **headers)
#         print(response)

#         user = User.objects.get(pk=1)
#         print(user.profile.info)


#     def test_is_active(self):
#         # login as second user to get token
#         response = self.client.post(self.login_url, self.user1_params)
#         token = response.json()["key"]
#         headers = {"HTTP_AUTHORIZATION": "Token " + token}

#         new_data = {"action": "check", "checkoutId": "206702119-chre59f1e77e6ca-132bd39c55"}
#         response = self.client.post(self.subscription_url, new_data, **headers)

#         user = User.objects.get(pk=1)
#         print(user.profile.info)

#         new_data = {"action": "is_active"}
#         response = self.client.post(self.subscription_url, new_data, **headers)
#         print(response)
#         print(response.json())

#         user = User.objects.get(pk=1)
#         print(user.profile.info)



import os
from datetime import datetime

import requests
from allauth.account.admin import EmailAddress
from django.conf import settings
from django.contrib.auth.models import User
from django.test import LiveServerTestCase
from django.utils.timezone import make_aware

from apps.accounts.models import Site
from apps.nbworker.rest import RESTClient
from apps.notebooks.models import Notebook
from apps.storage.models import UploadedFile
from apps.storage.s3utils import S3
from apps.storage.storage import StorageManager
from apps.workers.models import Worker, WorkerState

# python manage.py test apps.nbworker -v 2


class RESTClientTestCase(LiveServerTestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="developer",
            email="developer@example.com",
            password="developer",
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        self.site = Site.objects.create(
            title="Mercury",
            slug="single-site",
            share=Site.PUBLIC,
            created_by=self.user,
        )

        self.nb = Notebook.objects.create(
            title="some",
            slug="some",
            path="some",
            created_by=self.user,
            hosted_on=self.site,
            file_updated_at=make_aware(datetime.now()),
        )

    def test_provision_uploaded_files(self):
        settings.STORAGE = settings.STORAGE_S3
        os.environ["MERCURY_SERVER_URL"] = self.live_server_url
        session_id = "some-string"

        worker = Worker.objects.create(session_id=session_id, notebook=self.nb)
        sm = StorageManager(session_id, worker.id, self.nb.id)

        fname = "test.txt"
        with open(fname, "w") as fout:
            fout.write("test")

        s3 = S3()
        bucket_key = f"test/{fname}"
        s3.upload_file(fname, bucket_key)

        f = UploadedFile.objects.create(
            filename=fname,
            filepath=bucket_key,
            filetype="txt",
            filesize=5,
            hosted_on=self.site,
            created_by=self.user,
        )

        os.remove(fname)
        self.assertFalse(os.path.exists(fname))

        sm.provision_uploaded_files()

        self.assertTrue(os.path.exists(fname))

    def test_sync_output_dir(self):
        settings.STORAGE = settings.STORAGE_S3
        os.environ["MERCURY_SERVER_URL"] = self.live_server_url
        session_id = "some-string"
        worker = Worker.objects.create(session_id=session_id, notebook=self.nb)
        sm = StorageManager(session_id, worker.id, self.nb.id)

        output_dir = sm.worker_output_dir()
        fname = "test.txt"
        fpath = os.path.join(output_dir, "test.txt")
        print(output_dir, fpath, fname)
        with open(fpath, "w") as fout:
            fout.write("test")

        sm.sync_output_dir()

        url = f"{self.live_server_url}/api/v1/worker-output-files/{session_id}/{worker.id}/{self.nb.id}"
        response = requests.get(url)
        print(response)
        print(response.json())

    def test_save_nb_html(self):
        settings.STORAGE = settings.STORAGE_S3
        os.environ["MERCURY_SERVER_URL"] = self.live_server_url

        nb_html = "<html>test</html>"
        session_id = "some-string"
        worker = Worker.objects.create(session_id=session_id, notebook=self.nb)

        sm = StorageManager(session_id, worker.id, self.nb.id)
        html_path, html_url = sm.save_nb_html(nb_html)
        response = requests.get(html_url)
        self.assertTrue(response.text, nb_html)

    def test_get_nb(self):
        session_id = "some-string"
        worker = Worker.objects.create(session_id=session_id, notebook=self.nb)

        os.environ["MERCURY_SERVER_URL"] = self.live_server_url
        client = RESTClient(self.nb.id, session_id, worker.id)
        client.load_notebook()
        self.assertTrue(client.notebook.id, self.nb.id)

        with self.assertRaises(SystemExit):
            wrong_nb_id = 2
            client = RESTClient(wrong_nb_id, session_id, worker.id)
            client.load_notebook()

        with self.assertRaises(SystemExit):
            wrong_session = "some-wrong-session"
            client = RESTClient(wrong_nb_id, wrong_session, worker.id)
            client.load_notebook()

    def test_save_worker_state(self):
        session_id = "some-string"
        worker = Worker.objects.create(
            session_id=session_id, notebook=self.nb, state=WorkerState.Unknown
        )

        os.environ["MERCURY_SERVER_URL"] = self.live_server_url
        client = RESTClient(self.nb.id, session_id, worker.id)

        client.set_worker_state(new_state=WorkerState.Busy)
        worker = Worker.objects.get(pk=worker.id)
        self.assertTrue(worker.state, WorkerState.Busy)



import logging
import os
import subprocess
import sys
import requests

from celery import shared_task
from django.conf import settings

from apps.workers.models import Worker
from apps.ws.utils import machine_uuid

log = logging.getLogger(__name__)


@shared_task(bind=True)
def task_start_websocket_worker(self, job_params):
    log.debug(f"NbWorkers per machine: {settings.NBWORKERS_PER_MACHINE}")

    workers_ip_list = os.environ.get("WORKERS_IP_LIST", "")

    if workers_ip_list == "":
        machine_id = machine_uuid()
        workers = Worker.objects.filter(machine_id=machine_id)

        log.debug(f"Workers count: {len(workers)} machine_id={ machine_id }")

        if len(workers) > settings.NBWORKERS_PER_MACHINE:
            log.debug("Defer task start ws worker")
            task_start_websocket_worker.s(job_params).apply_async(countdown=15)
        else:
            directory = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            command = [
                sys.executable,
                os.path.join(directory, "nbworker"),
                str(job_params["notebook_id"]),
                str(job_params["session_id"]),
                str(job_params["worker_id"]),
                job_params["server_url"],
            ]
            log.debug("Start " + " ".join(command))
            worker = subprocess.Popen(command)
    else:
        workers_ips = workers_ip_list.split(",")
        print(workers_ips)

        all_busy = True

        for worker_ip in workers_ips:
            try:
                workers = Worker.objects.filter(machine_id=worker_ip)

                log.debug(f"Workers count: {len(workers)} machine_id={ worker_ip }")

                if len(workers) <= settings.NBWORKERS_PER_MACHINE:
                    notebook_id = job_params["notebook_id"]
                    session_id = job_params["session_id"]
                    worker_id = job_params["worker_id"]
                    response = requests.get(
                        f"http://{worker_ip}/start/{notebook_id}/{session_id}/{worker_id}"
                    )
                    if response.status_code == 200:
                        if response.json().get("msg", "") == "ok":
                            all_busy = False
                            break
            except Exception as e:
                pass

        if all_busy:
            log.debug("Defer task start ws worker")
            task_start_websocket_worker.s(job_params).apply_async(countdown=15)



from http.client import FORBIDDEN
import json

from django.db import transaction
from django.db.models import Q
from rest_framework import permissions, status, viewsets
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.serializers import ValidationError

from apps.accounts.models import Membership, Site
from apps.accounts.serializers import SiteSerializer
from apps.accounts.tasks import task_init_site
from apps.accounts.views.utils import get_slug, is_cloud_version
from apps.accounts.views.permissions import HasEditRights

PLAN_KEY = "plan"
PLAN_STARTER = "starter"
PLAN_PRO = "pro"
PLAN_BUSINESS = "business"

FORBIDDEN_SLUGS = [
    "mercury",
    "dashboard",
    "report",
    "mljar",
    "cloud",
    "api",
    "python",
    "backend",
    "frontend",
    "www",
    "app",
]


def get_plan(user):
    info = json.loads(user.profile.info)
    user_plan = info.get(PLAN_KEY, PLAN_STARTER)
    if user_plan not in [PLAN_STARTER, PLAN_PRO, PLAN_BUSINESS]:
        user_plan = PLAN_STARTER
    return user_plan


def max_number_of_sites(user):
    if not is_cloud_version():
        return 1000
    try:
        user_plan = get_plan(user)
        sites_plans = {PLAN_STARTER: 1, PLAN_PRO: 3, PLAN_BUSINESS: 10}
        return sites_plans[user_plan]
    except Exception as e:
        pass
    return 1


class SiteViewSet(viewsets.ModelViewSet):
    serializer_class = SiteSerializer
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        return Site.objects.filter(
            Q(
                pk__in=Membership.objects.filter(
                    user=self.request.user, rights=Membership.EDIT
                ).values("host__id")
            )
            | Q(created_by=self.request.user)
        )

    def create(self, request, *args, **kwargs):
        # check number of allowed sites
        if Site.objects.filter(created_by=request.user).count() >= max_number_of_sites(
            request.user
        ):
            return Response(
                {"msg": "Sites limit reached. Please change subsription plan"},
                status=status.HTTP_403_FORBIDDEN,
            )
        proposed_slug = get_slug(
            request.data.get("slug", ""), request.data.get("title", "")
        )
        if proposed_slug in FORBIDDEN_SLUGS:
            return Response(
                {"msg": "Please change site subdomain, current value is forbidden"},
                status=status.HTTP_403_FORBIDDEN,
            )
        if Site.objects.filter(slug=proposed_slug):
            return Response(
                {"msg": "Please change site subdomain, current value is not unique"},
                status=status.HTTP_403_FORBIDDEN,
            )

        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def perform_create(self, serializer):
        try:
            # create a database instance
            with transaction.atomic():
                instance = serializer.save(created_by=self.request.user)

                instance.slug = get_slug(instance.slug, instance.title)

                instance.save()
        except Exception as e:
            raise APIException(str(e))

    def perform_update(self, serializer):
        updated_instance = serializer.save()
        # lets check slug if we update it
        new_slug = self.request.data.get("slug")

        if new_slug is not None:
            new_slug = get_slug(new_slug, updated_instance.title)
            if new_slug in FORBIDDEN_SLUGS:
                raise ValidationError(f"You cant use {new_slug} as a subdomain")
            updated_instance.slug = new_slug
            updated_instance.save()

    def destroy(self, request, *args, **kwargs):
        """Only owner can delete object"""
        try:
            instance = self.get_object()
            if instance.created_by == self.request.user:
                self.perform_destroy(instance)
            else:
                return Response(status=status.HTTP_404_NOT_FOUND)
        except Exception:
            return Response(status=status.HTTP_404_NOT_FOUND)
        return Response(status=status.HTTP_204_NO_CONTENT)


class GetSiteView(APIView):
    def get(self, request, site_slug, format=None):
        url = site_slug
        # this can be a custom domain
        custom_domain = url
        # or subdomain with domain
        subdomain, domain = "single-site", "runmercury.com"

        if len(url.split(".")) > 1:
            subdomain = url.split(".")[0]
            domain = ".".join(url.split(".")[1:])

        if site_slug in ["127.0.0.1", "localhost"]:
            subdomain = "single-site"
            domain = "runmercury.com"

        if request.build_absolute_uri().startswith("http://127.0.0.1"):
            subdomain = "single-site"
            domain = "runmercury.com"

        # print(f"{url}|{request.build_absolute_uri()}|{subdomain}|{domain}|{custom_domain}")

        sites = Site.objects.filter(
            Q(custom_domain=custom_domain) | Q(slug=subdomain, domain=domain)
        )
        if not sites:
            return Response(status=status.HTTP_404_NOT_FOUND)

        if sites[0].share == Site.PUBLIC:
            return Response(SiteSerializer(sites[0]).data)

        if request.user.is_anonymous:
            return Response(status=status.HTTP_403_FORBIDDEN)

        sites = sites.filter(
            # any Membership (VIEW or EDIT) or owner
            Q(
                pk__in=Membership.objects.filter(user=self.request.user).values(
                    "host__id"
                )
            )
            | Q(created_by=self.request.user)
        )

        if not sites:
            return Response(status=status.HTTP_403_FORBIDDEN)

        return Response(SiteSerializer(sites[0]).data)


class InitializeSite(APIView):
    permission_classes = [permissions.IsAuthenticated, HasEditRights]

    def post(self, request, site_id, format=None):
        try:
            with transaction.atomic():
                job_params = {"site_id": site_id}
                transaction.on_commit(lambda: task_init_site.delay(job_params))
                return Response(status=status.HTTP_200_OK)
        except Exception as e:
            raise APIException(str(e))



import json

import ipywidgets
from IPython.display import display

from .manager import WidgetsManager


class MultiSelect:
    def __init__(
        self, value=[], choices=[], label="", url_key="", disabled=False, hidden=False
    ):
        if value is None and len(choices) > 1:
            value = [choices[0]]

        self.code_uid = WidgetsManager.get_code_uid("MultiSelect", key=url_key)
        self.url_key = url_key
        self.hidden = hidden
        choices = list(choices)
        value = list(value)
        if WidgetsManager.widget_exists(self.code_uid):
            self.select = WidgetsManager.get_widget(self.code_uid)
            if list(self.select.options) != choices:
                self.select.options = choices
                self.select.value = value
            self.select.description = label
            self.select.disabled = disabled
        else:
            self.select = ipywidgets.SelectMultiple(
                value=value,
                options=choices,
                description=label,
                style={"description_width": "initial"},
                disabled=disabled,
            )
            WidgetsManager.add_widget(self.select.model_id, self.code_uid, self.select)
        display(self)

    @property
    def value(self):
        return list(self.select.value)

    @value.setter
    def value(self, v):
        try:
            self.select.value = v
        except Exception as e:
            if len(self.select.options) > 0:
                self.select.value = self.select.options[0]
            else:
                self.select.value = None

    def __str__(self):
        return "m.MultiSelect"

    def __repr__(self):
        return "mercury.MultiSelect"

    def _repr_mimebundle_(self, **kwargs):
        # data = {}
        # data["text/plain"] = repr(self)
        # return data
        data = self.select._repr_mimebundle_()

        if len(data) > 1:
            view = {
                "widget": "MultiSelect",
                "value": self.select.value,
                "choices": self.select.options,
                "label": self.select.description,
                "model_id": self.select.model_id,
                "code_uid": self.code_uid,
                "url_key": self.url_key,
                "disabled": self.select.disabled,
                "hidden": self.hidden,
            }
            data["application/mercury+json"] = json.dumps(view, indent=4)
            if "text/plain" in data:
                del data["text/plain"]

            if self.hidden:
                key = "application/vnd.jupyter.widget-view+json"
                if key in data:
                    del data[key]

            return data



from rest_framework import serializers

from apps.workers.models import Worker


class WorkerSerializer(serializers.ModelSerializer):
    class Meta:
        model = Worker
        read_only_fields = ("id", "created_at", "updated_at")
        fields = ("id", "created_at", "updated_at", "state")



import os

from django.core.management.base import BaseCommand, CommandError

from apps.notebooks.models import Notebook


class Command(BaseCommand):
    help = "Delete a new notebook"

    def add_arguments(self, parser):
        parser.add_argument("notebook_path", help="Path to notebook")

    def handle(self, *args, **options):
        notebook_path = options["notebook_path"]
        notebook_path = os.path.abspath(notebook_path)
        self.stdout.write(self.style.HTTP_INFO(f"Try to delete {notebook_path}"))
        notebooks_deleted_cnt, _ = Notebook.objects.filter(path=notebook_path).delete()
        if notebooks_deleted_cnt:
            self.stdout.write(self.style.SUCCESS("Successfully deleted a notebook"))
        else:
            self.stdout.write(
                self.style.NOTICE("No notebooks with provided path deleted")
            )



import os
import subprocess
import sys
import time

import psutil
from django.core.management.base import BaseCommand, CommandError

from apps.notebooks.models import Notebook
from apps.notebooks.tasks import task_init_notebook


def kill(proc_pid):
    try:
        process = psutil.Process(proc_pid)
        for proc in process.children(recursive=True):
            proc.kill()
        process.kill()
    except Exception as e:
        pass


class Command(BaseCommand):
    help = "Watch notebook"

    def add_arguments(self, parser):
        parser.add_argument("notebook_path", help="Path to the notebook")

    def handle(self, *args, **options):
        try:
            self.stdout.write(
                self.style.HTTP_INFO(f'Watching notebook {options["notebook_path"]}')
            )

            notebook_id = task_init_notebook(
                options["notebook_path"], is_watch_mode=True
            )

            logo = """                                                                                  
     _ __ ___   ___ _ __ ___ _   _ _ __ _   _ 
    | '_ ` _ \ / _ \ '__/ __| | | | '__| | | |
    | | | | | |  __/ | | (__| |_| | |  | |_| |
    |_| |_| |_|\___|_|  \___|\__,_|_|   \__, |
                                         __/ |
                                        |___/ 
            """

            self.stdout.write(self.style.SUCCESS("-" * 53))
            self.stdout.write(self.style.SUCCESS(logo))
            self.stdout.write(self.style.SUCCESS("-" * 53))
            self.stdout.write(
                self.style.SUCCESS(
                    f"Please open the following address in your web browser"
                )
            )
            self.stdout.write(
                self.style.SUCCESS(f"--> http://127.0.0.1:8000/app/{notebook_id}")
            )
            self.stdout.write(self.style.SUCCESS("-" * 53))

            mercury_bin = sys.argv[0]
            # check if we are on Windows machines and
            # need to extend the mercury bin with .exe
            if sys.executable.endswith(".exe") and not mercury_bin.endswith(".exe"):
                mercury_bin += ".exe"

            server_command = [
                sys.executable,
                mercury_bin,
                "runserver",
                "--noreload",
                "--noadditional",
            ]
            server = subprocess.Popen(server_command)

            worker_command = [
                "celery",
                "-A",
                "mercury.server" if sys.argv[0].endswith("mercury") else "server",
                "worker",
                "--loglevel=error",
                "-P",
                "gevent",
                "--concurrency",
                "1",
                "-E",
            ]
            worker = subprocess.Popen(worker_command)

            server.wait()
            worker.wait()

        except KeyboardInterrupt:
            try:
                kill(server.pid)
                kill(worker.pid)
                self.delete_notebook(notebook_id)
                self.clear_celery_backend()
                self.stdout.write(self.style.SUCCESS("Stop watching"))
                sys.exit(0)
            except SystemExit:
                os._exit(0)
        except Exception as e:
            print("Mercury watch error.", str(e))

    def delete_notebook(self, id):
        Notebook.objects.filter(pk=id).delete()
        self.stdout.write(self.style.HTTP_INFO(f"Notebook deleted"))

    def clear_celery_backend(self):
        try:
            if os.path.exists("celery.sqlite"):
                os.remove("celery.sqlite")
        except Exception as e:
            print("Problem with removing Celery backend", str(e))



# Please run tests with below command
# python manage.py test apps.accounts.tests.test_accounts

from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core import mail
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Membership, Secret, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker


class AccountsTestCase(APITestCase):
    register_url = "/api/v1/auth/register/"
    verify_email_url = "/api/v1/auth/register/verify-email/"
    login_url = "/api/v1/auth/login/"
    user_details_url = "/api/v1/auth/user/"
    logout_url = "/api/v1/auth/logout/"
    delete_account_url = "/api/v1/auth/delete-account/"
    change_password_url = "/api/v1/auth/password/change/"
    reset_password_url = "/api/v1/auth/password/reset/"
    reset_password_confirm_url = "/api/v1/auth/password/reset/confirm/"

    sites_url = "/api/v1/sites/"
    members_url = "/api/v1/{}/members/"  # need to set site_id
    invite_url = "/api/v1/{}/invite/"

    def setUp(self):
        # os.environ["ACCOUNT_EMAIL_VERIFICATION"] = "mandatory"
        #
        # first user
        #
        self.user1_params = {
            "username": "user1",  # it is optional to pass username
            "email": "piotr@example.com",
            "password": "verysecret",
        }
        self.user = User.objects.create_user(
            username=self.user1_params["username"],
            email=self.user1_params["email"],
            password=self.user1_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        #
        # second user
        #
        self.user2_params = {
            "username": "user2",  # it is optional to pass username
            "email": "piotr2@example.com",
            "password": "verysecret2",
        }
        self.user2 = User.objects.create_user(
            username=self.user2_params["username"],
            email=self.user2_params["email"],
            password=self.user2_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user2, email=self.user2.email, verified=True, primary=True
        )

    def test_list_members(self):
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )
        # login as first user to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        response = self.client.get(self.members_url.format(site.id), **headers)
        self.assertEqual(len(response.json()), 0)

        member = Membership.objects.create(
            user=self.user2, host=site, rights=Membership.VIEW, created_by=self.user
        )

        response = self.client.get(self.members_url.format(site.id), **headers)
        self.assertEqual(len(response.json()), 1)

        # login as second user
        response = self.client.post(self.login_url, self.user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        response = self.client.get(self.members_url.format(site.id), **headers)
        # no members because second user has VIEW rights
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json()["detail"],
            "You do not have permission to perform this action.",
        )
        # change the rights
        member.rights = Membership.EDIT
        member.save()
        # we get the list
        response = self.client.get(self.members_url.format(site.id), **headers)
        # no members because second user has VIEW rights
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()), 1)

    def test_create_member(self):
        site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )
        # login as first user to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"user_id": 2, "rights": Membership.VIEW}
        response = self.client.post(
            self.members_url.format(site.id), new_data, **headers
        )
        self.assertEqual(response.status_code, 201)
        # example response
        # {'id': 1, 'created_at': '2023-02-23T14:16:26.210009Z', 'created_by': 1, 'updated_at': '2023-02-23T14:16:26.210310Z', 'rights': 'VIEW', 'user': {'pk': 2, 'username': 'user2', 'email': 'piotr2@example.com', 'first_name': '', 'last_name': ''}}

    def test_register(self):
        # register data
        data = {
            "email": "user2@example-email.com",
            "password1": "verysecret",
            "password2": "verysecret",
        }
        # send POST request to "/api/auth/register/"
        response = self.client.post(self.register_url, data)
        # check the response status and data
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()["detail"], "Verification e-mail sent.")

        # try to login - should fail, because email is not verified
        login_data = {
            "email": data["email"],
            "password": data["password1"],
        }
        response = self.client.post(self.login_url, login_data)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertTrue(
            "E-mail is not verified." in response.json()["non_field_errors"]
        )

        # expected one email to be send
        # parse email to get token
        self.assertEqual(len(mail.outbox), 1)
        email_lines = mail.outbox[0].body.splitlines()
        activation_line = [l for l in email_lines if "verify-email" in l][0]
        activation_link = activation_line.split("go to ")[1]
        activation_key = activation_link.split("/")[4]

        response = self.client.post(self.verify_email_url, {"key": activation_key})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()["detail"], "ok")

        # lets login after verification to get token key
        response = self.client.post(self.login_url, login_data)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertTrue("key" in response.json())

    def test_get_user_details(self):
        # login to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        # set headers
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        # get user details
        response = self.client.get(self.user_details_url, **headers)
        self.assertEqual(response.status_code, 200)
        data = response.json()
        for k in ["username", "email", "profile"]:
            self.assertTrue(k in data)
        self.assertTrue("info" in data["profile"])



import os
import json
import logging
import platform
import requests
import nbformat

from widgets.manager import WidgetsManager

log = logging.getLogger(__name__)


CLIENT_SITE = "client"
WORKER_SITE = "worker"


my_ip = None


def machine_uuid():
    global my_ip
    if my_ip is not None:
        return my_ip
    if os.environ.get("USE_WORKER_IP") is not None:
        response = requests.get("https://api.ipify.org?format=json")
        my_ip = response.json().get("ip", platform.node())
        return my_ip
    return platform.node()


def client_group(notebook_id, session_id):
    return f"{CLIENT_SITE}-{notebook_id}-{session_id}"


def worker_group(notebook_id, session_id):
    return f"{WORKER_SITE}-{notebook_id}-{session_id}"


def is_presentation(nb):
    for cell in nb["cells"]:
        if "slideshow" in cell.get("metadata", {}):
            return True
    return False


def parse_params(nb, params={}):
    # nb in nbformat

    all_model_ids = []

    cell_counter = 1

    no_outputs = True

    mercury_package_imported = False

    for cell in nb["cells"]:
        if not mercury_package_imported:
            # try to guess if mercury is imported
            code = cell.get("source", "")
            if isinstance(code, list):
                code = "".join(code)
            if "mercury" in code:
                mercury_package_imported = True

        for output in cell.get("outputs", []):
            if "data" in output and "application/mercury+json" in output["data"]:
                no_outputs = False
                view = output["data"]["application/mercury+json"]

                if isinstance(view, nbformat.notebooknode.NotebookNode):
                    view = view.dict()
                else:
                    view = json.loads(str(view))

                # print(
                #    f'model_id={view.get("model_id", "")} in all models ids {all_model_ids}'
                # )

                # check model_id duplicates
                if view.get("model_id", "") in all_model_ids:
                    continue

                widget_type = view.get("widget")
                if widget_type is None:
                    continue
                else:
                    params["version"] = "2"
                    if "params" not in params:
                        params["params"] = {}

                widget_key = view.get("code_uid")
                if widget_key is None:
                    continue
                # fix cell index, cell index is not set in Jupyter Notebook mode
                widget_key = WidgetsManager.fix_cell_index(widget_key, cell_counter)

                if widget_type == "App":
                    if view.get("title") is not None:
                        params["title"] = view.get("title")
                    if view.get("description") is not None:
                        params["description"] = view.get("description")
                    if view.get("show_code") is not None:
                        params["show-code"] = view.get("show_code")
                    if view.get("show_prompt") is not None:
                        params["show-prompt"] = view.get("show_prompt")
                    if view.get("output") is not None:
                        params["output"] = view.get("output")
                    if view.get("schedule") is not None:
                        params["schedule"] = view.get("schedule")
                    if view.get("notify") is not None:
                        params["notify"] = json.loads(view.get("notify"))
                    if view.get("continuous_update") is not None:
                        params["continuous_update"] = view.get("continuous_update")
                    if view.get("static_notebook") is not None:
                        params["static_notebook"] = view.get("static_notebook")

                    for property in ["show_sidebar", "full_screen", "allow_download"]:
                        if view.get(property) is not None:
                            params[property] = view.get(property)

                else:
                    params["params"][widget_key] = WidgetsManager.frontend_format(view)

                all_model_ids += [view.get("model_id", "")]

        cell_counter += 1

    if params.get("show-code") is None:
        params["show-code"] = False
    if params.get("show-prompt") is None:
        params["show-prompt"] = True
    if params.get("continuous_update") is None:
        params["continuous_update"] = True
    if params.get("static_notebook") is None:
        params["static_notebook"] = not mercury_package_imported
    if params.get("show_sidebar") is None:
        params["show_sidebar"] = True
    if params.get("full_screen") is None:
        params["full_screen"] = True
    if params.get("allow_download") is None:
        params["allow_download"] = True

    if no_outputs:
        params["version"] = "2"
        params["show-code"] = False
        params["show-prompt"] = False
        params["params"] = {}
        params["output"] = "app"

    if is_presentation(nb):
        params["output"] = "slides"



# Please run tests with below command
# python manage.py test apps.accounts

from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core import mail
from django.utils.timezone import make_aware
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Invitation, Membership, Site
from apps.accounts.tasks import task_send_invitation, task_send_new_member
from apps.notebooks.models import Notebook
from apps.workers.models import Worker


class InvitationsTestCase(APITestCase):
    register_url = "/api/v1/auth/register/"
    login_url = "/api/v1/auth/login/"
    invite_url = "/api/v1/{}/invite"
    list_invitations_url = "/api/v1/{}/list-invitations"
    delete_invitation_url = "/api/v1/{}/delete-invitation/{}"

    def setUp(self):
        self.user1_params = {
            "username": "user1",  # it is optional to pass username
            "email": "piotr@example.com",
            "password": "verysecret",
        }
        self.user = User.objects.create_user(
            username=self.user1_params["username"],
            email=self.user1_params["email"],
            password=self.user1_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        self.site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )

    def test_send_invitation(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        new_data = {"email": "some@example.com", "rights": "EDIT"}

        response = self.client.post(
            self.invite_url.format(self.site.id), new_data, **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(Invitation.objects.all()), 1)

        task_send_invitation({"invitation_id": 1})

        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue("invites" in mail.outbox[0].body)
        self.assertTrue("edit" in mail.outbox[0].body)
        self.assertTrue(new_data["email"] in mail.outbox[0].to)

        new_data = {
            "email": new_data["email"],
            "password1": "verysecret",
            "password2": "verysecret",
        }
        response = self.client.post(self.register_url, new_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

    def test_invite_already_registered_user(self):
        new_email = "some@example.com"
        new_data = {
            "email": new_email,
            "password1": "verysecret",
            "password2": "verysecret",
        }
        response = self.client.post(self.register_url, new_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        # create invitation
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"email": new_email, "rights": "EDIT"}
        response = self.client.post(
            self.invite_url.format(self.site.id), new_data, **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(Invitation.objects.all()), 0)

        self.assertEqual(len(Membership.objects.all()), 1)

        task_send_new_member({"membership_id": 1})

        # first email is from registration
        # second email is invitations
        self.assertEqual(len(mail.outbox), 2)
        self.assertTrue("invites" in mail.outbox[1].body)
        self.assertTrue("edit" in mail.outbox[1].body)
        self.assertTrue(new_email in mail.outbox[1].to)

    def test_list_invitations(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        # create invitation
        new_email = "some@example.com"
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"email": new_email, "rights": "VIEW"}
        response = self.client.post(
            self.invite_url.format(self.site.id), new_data, **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(Invitation.objects.all()), 1)

        response = self.client.get(
            self.list_invitations_url.format(self.site.id), **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json()[0]["invited"], new_email)

    def test_delete_invitation(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        # create invitation
        new_email = "some@example.com"
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"email": new_email, "rights": "VIEW"}
        response = self.client.post(
            self.invite_url.format(self.site.id), new_data, **headers
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(Invitation.objects.all()), 1)

        response = self.client.delete(
            self.delete_invitation_url.format(self.site.id, 1), **headers
        )
        self.assertEqual(response.status_code, 204)
        self.assertEqual(response.status_code, 204)



import os

from django.core.management.base import BaseCommand, CommandError

from apps.notebooks.models import Notebook


class Command(BaseCommand):
    help = "List all notebooks"

    def handle(self, *args, **options):
        notebooks = Notebook.objects.all()
        if not notebooks:
            self.stdout.write(self.style.NOTICE("No notebooks available"))
        else:
            self.stdout.write(self.style.SUCCESS("Notebooks list:"))
            for n in notebooks:
                print(f"(id:{n.id}) {n.title} from {n.path}")



from IPython.display import display, HTML


def Chat(messages=[]):
    html = """<div style="padding: 20px;">"""

    left_msg = """<div style="background: #efefef; padding: 10px; float: left; border-radius: 10px 10px 10px 0px; margin: 5px; margin-right: 50px">
<pre style="background: #efefef; color: black;">
{}
</pre>
</div>
<div style="clear: both;"></div>"""
    right_msg = """<div style="background: cornflowerblue; color: #fff; padding: 10px; float: right; border-radius: 10px 10px 0px 10px; margin: 5px; margin-left: 50px">
<pre style="background: cornflowerblue; color: #fff;">
{}
</pre>
</div>
<div style="clear: both;"></div>"""
    style = left_msg
    for m in messages:
        html += style.format(m)
        style = left_msg if style == right_msg else right_msg

    html += "</div>"

    display(HTML(html))



from django.conf import settings
from django.conf.urls import include
from django.conf.urls.static import static
from django.contrib import admin
from django.shortcuts import render
from django.urls import path, re_path

from apps.notebooks.urls import notebooks_urlpatterns
from apps.tasks.urls import tasks_urlpatterns
from apps.accounts.urls import accounts_urlpatterns
from apps.storage.urls import storage_urlpatterns
from apps.workers.urls import workers_urlpatterns
from server.views import VersionInfo, WelcomeMessage

urlpatterns = []

if settings.DEBUG or settings.SERVE_STATIC:
    # serve static file for development only!
    def index(request):
        return render(request, "index.html")

    # Serve static and media files from development server
    urlpatterns += [
        path("", index),
        re_path(r"^app", index),
        re_path(r"^login", index),
        re_path(r"^account", index),
    ]
    urlpatterns += static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)


urlpatterns += [
    path("admin/", admin.site.urls),
    re_path(
        "api/v1/version",
        VersionInfo.as_view(),
    ),
    re_path(
        "api/v1/(?P<site_id>.+)/welcome",
        WelcomeMessage.as_view(),
    ),
    re_path(r"^api/v1/fp/", include("django_drf_filepond.urls")),
]

urlpatterns += tasks_urlpatterns
urlpatterns += notebooks_urlpatterns
urlpatterns += accounts_urlpatterns
urlpatterns += storage_urlpatterns
urlpatterns += workers_urlpatterns

admin.site.site_header = "Mercury Admin Panel"



import logging

from django.db.models import Q
from django.http import JsonResponse
from django.conf import settings
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from apps.accounts.models import Membership, Site
from apps.notebooks.models import Notebook
from apps.storage.models import UploadedFile, WorkerFile
from apps.storage.s3utils import S3
from apps.storage.serializers import UploadedFileSerializer
from apps.workers.models import Worker
from apps.accounts.views.utils import is_cloud_version

from apps.accounts.views.sites import (
    get_plan,
    PLAN_KEY,
    PLAN_STARTER,
    PLAN_PRO,
    PLAN_BUSINESS,
)

from apps.storage.utils import (
    get_bucket_key,
    get_worker_bucket_key,
)

log = logging.getLogger(__name__)


class GetStorageType(APIView):
    def get(self, request, format=None):
        return Response({"storage_type": settings.STORAGE})


def get_site(user, site_id):
    if user.is_anonymous:
        return None

    sites = Site.objects.filter(
        Q(pk=site_id)
        & (
            Q(
                pk__in=Membership.objects.filter(
                    user=user, rights=Membership.EDIT
                ).values("host__id")
            )
            | Q(created_by=user)
        )
    )

    if not sites:
        return None
    return sites[0]


class ListFiles(APIView):
    def get(self, request, site_id, format=None):
        site = get_site(request.user, site_id)
        if site is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        files = UploadedFile.objects.filter(hosted_on=site)

        return Response(UploadedFileSerializer(files, many=True).data)


class PresignedUrl(APIView):
    def get(self, request, action, site_id, filename, format=None):
        site = get_site(request.user, site_id)
        if site is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        client_action = (
            "put_object" if action in ["put", "put_object"] else "get_object"
        )

        s3 = S3()
        url = s3.get_presigned_url(
            get_bucket_key(site, request.user, filename.replace(" ", "-")),
            client_action,
        )
        return Response({"url": url})


FILE_LIMITS = {
    PLAN_STARTER: {
        "files": 10,
        "size": 10,
    },
    PLAN_PRO: {"files": 25, "size": 50},
    PLAN_BUSINESS: {"files": 50, "size": 100},  # MB
}


def upload_allowed_check_limits(user, site_id, filesize):
    if not is_cloud_version():
        return True
    plan = get_plan(user)
    files_count_limit = FILE_LIMITS[plan]["files"]
    files_size_limit = FILE_LIMITS[plan]["size"]  # in MB

    if int(filesize) / 1024 / 1024 > files_size_limit:
        return False

    total_files = UploadedFile.objects.filter(hosted_on__id=site_id)
    if total_files.count() > files_count_limit:
        return False

    return True


class PresignedUrlPut(APIView):
    def get(self, request, site_id, filename, filesize, format=None):
        site = get_site(request.user, site_id)
        if site is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        upload_allowed = upload_allowed_check_limits(request.user, site_id, filesize)
        if not upload_allowed:
            return Response(status=status.HTTP_403_FORBIDDEN)

        client_action = "put_object"
        s3 = S3()
        url = s3.get_presigned_url(
            get_bucket_key(site, request.user, filename.replace(" ", "-")),
            client_action,
        )
        return Response({"url": url})


class WorkerPresignedUrl(APIView):
    def get(
        self,
        request,
        action,
        session_id,
        worker_id,
        notebook_id,
        output_dir,
        filename,
        format=None,
    ):
        try:
            # check if such worker exists
            Worker.objects.get(
                pk=worker_id, session_id=session_id, notebook__id=notebook_id
            )
            client_action = (
                "put_object" if action in ["put", "put_object"] else "get_object"
            )

            s3 = S3()
            url = s3.get_presigned_url(
                get_worker_bucket_key(
                    session_id, output_dir, filename.replace(" ", "-")
                ),
                client_action,
            )

            return Response({"url": url})

        except Exception as e:
            log.exception("Cant create presigned url for worker")

        return Response(status=status.HTTP_403_FORBIDDEN)


class FileUploaded(APIView):
    def post(self, request, format=None):
        site_id = request.data.get("site_id")
        filename = request.data.get("filename", "").replace(" ", "-")
        filesize = request.data.get("filesize")
        filetype = filename.split(".")[-1].lower()

        site = get_site(request.user, site_id)
        if site is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        bucket_key = get_bucket_key(site, request.user, filename)

        # remove previous objects with the same filepath
        UploadedFile.objects.filter(filepath=bucket_key, hosted_on=site).delete()

        # create a new object
        UploadedFile.objects.create(
            filename=filename,
            filepath=bucket_key,
            filetype=filetype,
            filesize=filesize,
            hosted_on=site,
            created_by=request.user,
        )

        return Response(status=status.HTTP_200_OK)


class DeleteFile(APIView):
    def post(self, request, format=None):
        site_id = request.data.get("site_id")
        filename = request.data.get("filename", "").replace(" ", "-")

        site = get_site(request.user, site_id)
        if site is None:
            return Response(status=status.HTTP_403_FORBIDDEN)

        bucket_key = get_bucket_key(site, request.user, filename)

        s3 = S3()
        s3.delete_file(bucket_key)

        UploadedFile.objects.filter(filepath=bucket_key, hosted_on=site).delete()

        Notebook.objects.filter(path__icontains=filename, hosted_on=site).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)



# Please run tests with below command
# python manage.py test apps.accounts.tests.test_secrets

from datetime import datetime

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.utils.timezone import make_aware
from rest_framework.test import APITestCase

from apps.accounts.models import Membership, Secret, Site
from apps.notebooks.models import Notebook
from apps.workers.models import Worker


class SecretsTestCase(APITestCase):
    register_url = "/api/v1/auth/register/"
    login_url = "/api/v1/auth/login/"
    add_secret_url = "/api/v1/{}/add-secret"
    list_secrets_url = "/api/v1/{}/list-secrets"
    delete_secret_url = "/api/v1/{}/delete-secret/{}"

    def setUp(self):
        self.user1_params = {
            "username": "user1",  # it is optional to pass username
            "email": "piotr@example.com",
            "password": "verysecret",
        }
        self.user = User.objects.create_user(
            username=self.user1_params["username"],
            email=self.user1_params["email"],
            password=self.user1_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )
        self.site = Site.objects.create(
            title="First site", slug="first-site", created_by=self.user
        )

    def test_add_secret(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"name": "MY_SECRET", "secret": "super-secret"}
        response = self.client.post(
            self.add_secret_url.format(self.site.id), new_data, **headers
        )
        secrets = Secret.objects.all()

        self.assertEqual(len(secrets), 1)
        self.assertEqual(secrets[0].name, new_data["name"])
        self.assertNotEqual(secrets[0].token, new_data["secret"])

        response = self.client.get(
            self.list_secrets_url.format(self.site.id), **headers
        )
        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0]["name"], new_data["name"])

    def test_list_worker_secrets(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"name": "MY_SECRET", "secret": "super-secret"}
        response = self.client.post(
            self.add_secret_url.format(self.site.id), new_data, **headers
        )

        nb = Notebook.objects.create(
            title="some",
            slug="some",
            path="some",
            created_by=self.user,
            hosted_on=self.site,
            file_updated_at=make_aware(datetime.now()),
        )
        session_id = "some-session"
        worker = Worker.objects.create(
            session_id=session_id, machine_id="some-machine", notebook=nb
        )

        # get worker secrets
        url = f"/api/v1/worker/{session_id}/{worker.id}/{nb.id}/worker-secrets"
        response = self.client.get(url)
        self.assertEqual(response.json()[0].get("name"), new_data["name"])
        self.assertEqual(response.json()[0].get("secret"), new_data["secret"])

    def test_delete_secret(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"name": "MY_SECRET", "secret": "super-secret"}
        response = self.client.post(
            self.add_secret_url.format(self.site.id), new_data, **headers
        )
        secrets = Secret.objects.all()
        self.assertEqual(len(secrets), 1)

        response = self.client.delete(
            self.delete_secret_url.format(self.site.id, 1), **headers
        )
        self.assertEqual(response.status_code, 204)
        secrets = Secret.objects.all()
        self.assertEqual(len(secrets), 0)

    def test_list_secrets_by_user(self):
        # login
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}
        new_data = {"name": "MY_SECRET", "secret": "super-secret"}
        response = self.client.post(
            self.add_secret_url.format(self.site.id), new_data, **headers
        )

        # get worker secrets
        url = f"/api/v1/{self.site.id}/list-secrets"
        response = self.client.get(url, **headers)
        self.assertTrue(len(response.json()), 1)

        user2_params = {
            "username": "user2",  # it is optional to pass username
            "email": "piotr2@example.com",
            "password": "verysecret2",
        }
        user2 = User.objects.create_user(
            username=user2_params["username"],
            email=user2_params["email"],
            password=user2_params["password"],
        )
        EmailAddress.objects.create(
            user=user2, email=user2.email, verified=True, primary=True
        )

        response = self.client.post(self.login_url, user2_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        url = f"/api/v1/{self.site.id}/list-secrets"
        response = self.client.get(url, **headers)
        self.assertTrue(len(response.json()), 0)

        # invite
        m = Membership.objects.create(
            user=user2, host=self.site, rights=Membership.VIEW, created_by=self.user
        )

        url = f"/api/v1/{self.site.id}/list-secrets"
        response = self.client.get(url, **headers)
        self.assertTrue(len(response.json()), 0)

        m.rights = Membership.EDIT
        m.save()

        url = f"/api/v1/{self.site.id}/list-secrets"
        response = self.client.get(url, **headers)
        self.assertTrue(len(response.json()), 1)



from enum import Enum

from django.db import models

from apps.notebooks.models import Notebook


class WorkerState(str, Enum):
    Busy = "Busy"
    Running = "Running"
    Unknown = "Unknown"
    MaxRunTimeReached = "MaxRunTimeReached"
    MaxIdleTimeReached = "MaxIdleTimeReached"
    InstallPackages = "InstallPackages"


class Worker(models.Model):
    # machine unique id
    machine_id = models.CharField(max_length=128, blank=True)

    # web browser session id
    session_id = models.CharField(max_length=128)

    # notebook
    notebook = models.ForeignKey(
        Notebook,
        on_delete=models.CASCADE,
    )

    state = models.CharField(max_length=128, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)

    updated_at = models.DateTimeField(auto_now=True)



import os
import tempfile

import nbformat as nbf
from django.test import Client, TestCase
from execnb.nbio import dict2nb, read_nb

from apps.nb.exporter import Exporter
from apps.nb.nbrun import NbRun

# python manage.py test apps
from apps.nb.utils import one_cell_notebook, test_notebook
from apps.notebooks.models import Notebook
from apps.notebooks.tasks import task_init_notebook

# python manage.py test apps.nb


class NbTestCase(TestCase):
    def test_run_notebook(self):
        nb = test_notebook(code=["print(12)"])
        nb = dict2nb(nb)
        nbrun = NbRun()
        nbrun.run_notebook(nb)
        self.assertTrue("12" in "".join(nb.cells[0].outputs[0].text))

    def test_run_notebook(self):
        nb = test_notebook(code=["print(12)"])
        nb = dict2nb(nb)
        nbrun = NbRun()
        nbrun.run_notebook(nb)
        self.assertTrue("12" in "".join(nb.cells[0].outputs[0].text))

    def test_export_slides(self):
        nb = test_notebook(markdown=["# wow"])
        nb.cells[0]["metadata"] = {"slideshow": {"slide_type": "slide"}}
        nb = dict2nb(nb)
        nbrun = NbRun(is_presentation=True, reveal_theme="simple")
        nbrun.run_notebook(nb)
        body = nbrun.export_html(nb)
        self.assertTrue(len(body) > 0)



# Please run tests with below command
# python manage.py test apps.storage

import uuid

from allauth.account.admin import EmailAddress
from django.contrib.auth.models import User
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test.client import encode_multipart
from django_drf_filepond.models import TemporaryUpload
from rest_framework import status
from rest_framework.authtoken.models import Token
from rest_framework.test import APITestCase

from apps.accounts.models import Site
from apps.storage.s3utils import S3
from mercury.apps.storage.storage import StorageManager


class StorageTestCase(APITestCase):
    login_url = "/api/v1/auth/login/"
    upload_url = "/api/v1/fp/process/"

    def setUp(self):
        self.user1_params = {
            "username": "user1",  # it is optional to pass username
            "email": "piotr@example.com",
            "password": "verysecret",
        }
        self.user = User.objects.create_user(
            username=self.user1_params["username"],
            email=self.user1_params["email"],
            password=self.user1_params["password"],
        )
        EmailAddress.objects.create(
            user=self.user, email=self.user.email, verified=True, primary=True
        )

    def test_file_upload(self):
        fname = "test.txt"
        with open(fname, "w") as fout:
            fout.write("test file hello")

        data = None
        with open(fname, "rb") as fin:
            data = fin.read()

        file_spec = SimpleUploadedFile(fname, data)
        upload_form = {"filepond": file_spec}
        boundary = str(uuid.uuid4()).replace("-", "")

        encoded_form = encode_multipart(boundary, upload_form)
        content_type = "multipart/form-data; boundary=%s" % (boundary)

        # login as first user to get token
        response = self.client.post(self.login_url, self.user1_params)
        token = response.json()["key"]
        headers = {"HTTP_AUTHORIZATION": "Token " + token}

        print("file upload", data)
        new_data = {"filepond": data}
        response = self.client.post(
            self.upload_url, data=encoded_form, content_type=content_type, **headers
        )
        print(response)
        print(response.data)

        tu = TemporaryUpload.objects.get(upload_id=response.data)
        print(tu)
        # tu.delete()

    def test_s3_live(self):
        fname = "test-file.txt"
        with open(fname, "w") as fout:
            fout.write("hello there")

        s3 = S3()
        bucket_key = "my-folder/test.txt"
        s3.upload_file(fname, bucket_key)
        new_fname = "test-file-2.txt"
        s3.download_file(bucket_key, new_fname)

        with open(new_fname, "r") as fin:
            print(fin.read())

        print(s3.list_files("my-folder"))

        s3.delete_file(bucket_key)

        print(s3.list_files("my-folder"))

        url = s3.get_presigned_url(bucket_key, "put_object")
        print(url)

        content = None
        with open(new_fname, "rb") as fin:
            content = fin.read()

        import requests

        response = requests.put(url, content)
        print(response)
